"""
Common methods for transferring data to and from GRZ buckets.
"""

from typing import TYPE_CHECKING

import boto3
from boto3 import client as boto3_client  # type: ignore[import-untyped]
from botocore.config import Config as Boto3Config  # type: ignore[import-untyped]

if TYPE_CHECKING:
    from types_boto3_s3 import S3Client
    from types_boto3_s3.service_resource import S3ServiceResource
else:
    # avoid undefined objects when not type checking
    S3Client = object
    S3ServiceResource = object

from .models.s3 import S3Options


def _empty_str_to_none(string: str | None) -> str | None:
    # if user specifies empty strings, this might be an issue
    if string == "" or string is None:
        return None
    else:
        return string


def init_s3_client(s3_options: S3Options) -> S3Client:
    """Create a boto3 Client from a grz-cli configuration."""
    # configure proxies if proxy_url is defined
    proxy_url = s3_options.proxy_url
    s3_config = Boto3Config(
        proxies={"http": str(proxy_url), "https": str(proxy_url)} if proxy_url is not None else None,
        request_checksum_calculation=s3_options.request_checksum_calculation,
    )

    # Initialize S3 client for uploading
    s3_client: S3Client = boto3_client(
        service_name="s3",
        region_name=_empty_str_to_none(s3_options.region_name),
        api_version=_empty_str_to_none(s3_options.api_version),
        use_ssl=s3_options.use_ssl,
        endpoint_url=_empty_str_to_none(str(s3_options.endpoint_url)),
        aws_access_key_id=_empty_str_to_none(s3_options.access_key),
        aws_secret_access_key=_empty_str_to_none(s3_options.secret),
        aws_session_token=_empty_str_to_none(s3_options.session_token),
        config=s3_config,
    )

    return s3_client


def init_s3_resource(s3_options: S3Options) -> S3ServiceResource:
    """Create a boto3 Resource from a grz-cli configuration."""
    proxy_url = s3_options.proxy_url
    s3_config = Boto3Config(
        proxies={"http": str(proxy_url), "https": str(proxy_url)} if proxy_url is not None else None,
        request_checksum_calculation=s3_options.request_checksum_calculation,
    )
    s3_resource = boto3.resource(
        service_name="s3",
        region_name=_empty_str_to_none(s3_options.region_name),
        api_version=_empty_str_to_none(s3_options.api_version),
        use_ssl=s3_options.use_ssl,
        endpoint_url=_empty_str_to_none(str(s3_options.endpoint_url)),
        aws_access_key_id=_empty_str_to_none(s3_options.access_key),
        aws_secret_access_key=_empty_str_to_none(s3_options.secret),
        aws_session_token=_empty_str_to_none(s3_options.session_token),
        config=s3_config,
    )

    return s3_resource



from typing import Literal

from pydantic import AnyHttpUrl, AnyUrl

from .base import IgnoringBaseModel, IgnoringBaseSettings


class S3Options(IgnoringBaseModel):
    endpoint_url: AnyHttpUrl
    """
    The URL for the S3 service.
    """

    bucket: str
    """
    The name of the S3 bucket.
    """

    access_key: str | None = None
    """
    The access key for the S3 bucket.
    If undefined, it is read from the AWS_ACCESS_KEY_ID environment variable.
    """

    secret: str | None = None
    """
    The secret key for the S3 bucket.
    If undefined, it is read from the AWS_SECRET_ACCESS_KEY environment variable.
    """

    session_token: str | None = None
    """
    The session token for temporary credentials (optional).
    """

    region_name: str | None = None
    """
    The region name for the S3 bucket.
    """

    api_version: str | None = None
    """
    The S3 API version.
    """

    use_ssl: bool = True
    """
    Whether to use SSL for S3 operations.
    """

    proxy_url: AnyUrl | None = None
    """
    The proxy URL for S3 operations (optional).
    """

    request_checksum_calculation: Literal["when_supported", "when_required"] | None = None
    """
    Whether to calculate checksums for S3 request payloads (optional).
    Valid values are:

    * ``when_supported`` -- When set, a checksum will be calculated for
      all request payloads of operations modeled with the ``httpChecksum``
      trait where ``requestChecksumRequired`` is ``true`` or a
      ``requestAlgorithmMember`` is modeled.

    * ``when_required`` -- When set, a checksum will only be calculated
      for request payloads of operations modeled with the ``httpChecksum``
      trait where ``requestChecksumRequired`` is ``true`` or where a
      ``requestAlgorithmMember`` is modeled and supplied.

    Defaults to None.
    """

    multipart_chunksize: int = 256 * 1024**2
    """
    The size of the chunks to use for multipart uploads in bytes.
    """


class S3ConfigModel(IgnoringBaseSettings):
    s3: S3Options



"""
Validation methods for BAM files

e.g. warn on headers present, which can contain personally identifying
information
"""

import logging
from collections.abc import Generator
from os import PathLike

import pysam

log = logging.getLogger(__name__)


def validate_bam(bam_path: str | PathLike) -> Generator[str]:
    """Check if a BAM file is valid for submission"""
    errors: list[str] = []

    # disable pysam's SQ header presence enforcement (check_sq)
    #  these are used to optionally sort alignments by a custom contig order,
    #  which is unnecessary in unmapped BAM files
    bam_file = pysam.AlignmentFile(str(bam_path), mode="rb", check_sq=False)

    header = bam_file.header.to_dict()
    # ignore the HD key as it is almost always present and extremely
    # unlikely to accidentally contain identifying information
    concerning_keys = header.keys() - {"HD"}
    if concerning_keys:
        log.warning(f"Detected a header in BAM file '{bam_path}', ensure it contains no private information!")

    yield from errors



"""Module for tracking and logging the state of files over time."""

from __future__ import annotations

import copy
import json
import typing
from collections.abc import Callable
from os import PathLike
from pathlib import Path

from grz_pydantic_models.submission.metadata.v1 import File as SubmissionFileMetadata

from ..utils.io import read_multiple_json
from .states import State

Index = tuple[str, float, int]


class FileProgressLogger[T: State]:
    """
    A class to log and track the state of files over time. It stores file states in a log file of JSON entries
    and allows querying the state based on the file path and modification time.
    """

    _index = {"file_path": str, "modification_time": float, "size": int}
    # mapping of index -> (metadata, data)
    _file_states: dict[Index, tuple[SubmissionFileMetadata, T]]

    def __init__(self, log_file_path: str | PathLike):
        """
        Initializes the FileProgressLogger instance.

        :param fields: Dictionary of field names and their types, representing the state to track.
        :param log_file_path: Path to the JSON file where file progress will be logged.
        """
        self._file_path = Path(log_file_path)
        self._file_states = {}

        # Read existing file states from the log file
        self.read()

    def read(self):
        """
        Reads the log file and loads the file states into memory.

        :raises ValueError: If the path exists but is not a file.
        """
        if self._file_path.exists():
            if self._file_path.is_file():
                with open(self._file_path) as fd:
                    for row_dict in read_multiple_json(fd):
                        # Get index and cast them to the correct types
                        index = typing.cast(
                            Index,
                            tuple(self._index[k](row_dict[k]) for k in self._index),
                        )
                        # Get metadata
                        metadata = row_dict["metadata"]
                        # Get state
                        state = row_dict["state"]

                        self._file_states[index] = (metadata, state)
            else:
                raise ValueError(f"Path is not a file: '{str(self._file_path)}'")

    def cleanup(self, keep: list[tuple[PathLike, SubmissionFileMetadata]]):
        """
        Removes all entries from the log file and in-memory state that are not in the keep list.

        :param keep: List of tuples containing the file path and metadata of files to keep.
        """
        self._file_path.unlink(missing_ok=True)
        for file, file_metadata in keep:
            state = self.get_state(file, file_metadata)
            if state is not None:
                self.set_state(file, file_metadata, state)

    def _get_index(self, file_path: str | PathLike) -> Index:
        """
        Generates a unique index for a given file based on its name and modification time.

        :param file_path: Path object representing the file.
        :return: A tuple containing the file name and modification time.
        """
        file_path = Path(file_path).resolve()

        if file_path.is_file():
            return str(file_path), file_path.stat().st_mtime, file_path.stat().st_size
        else:
            return str(file_path), -1, -1  # catches files that do not exist

    # def get_index(self, file_path: Path, file_metadata: Dict) -> tuple:
    #     return self._get_index(file_path, file_metadata)

    def get_state(
        self,
        file_path: str | PathLike,
        file_metadata: dict | SubmissionFileMetadata,
        default: T | Callable[[Path, SubmissionFileMetadata], T] | None = None,
    ) -> T | None:
        """
        Retrieves the stored state of a file if it exists in the log.

        :param file_path: The file path to query for the state.
        :param file_metadata: The metadata of the file to query for the state.
        :param default: Default state to use if the file does not exist.
            Can be a Callable that takes the file path and the file metadata as input and returns some state:
            `Callable[[Path, SubmissionFileMetadata], T]`.

            The default state gets automatically saved as the state for this file in case there is no stored state.
        :return: A dictionary representing the file's state, or None if the file's state isn't logged.
        """
        file_path = Path(file_path)
        index = self._get_index(file_path)

        # get stored state
        stored_metadata, stored_data = self._file_states.get(index, (None, None))

        # check if metadata matches
        if not isinstance(file_metadata, SubmissionFileMetadata):
            file_metadata = SubmissionFileMetadata(**file_metadata)

        if stored_metadata and not isinstance(stored_metadata, SubmissionFileMetadata):
            stored_metadata = SubmissionFileMetadata(**stored_metadata)

        if stored_metadata and file_metadata == stored_metadata:
            return copy.deepcopy(stored_data)

        # metadata mismatch -> no valid stored state -
        if file_metadata and default:
            if callable(default):
                state = default(file_path, file_metadata)
                self.set_state(file_path, file_metadata, state)
                return state
            else:
                if not default:
                    raise ValueError("Default state must be provided if not callable")
                    # return None
                default = typing.cast(T, default)
                self.set_state(file_path, file_metadata, default)
                return default
        else:
            return None

    def set_state(
        self,
        file_path: str | PathLike,
        file_metadata: dict | SubmissionFileMetadata,
        state: T,
    ):
        """
        Log the state of a file:
         - Update the in-memory state
         - Persist the state to the JSON log file

        :param file_path: The path of the file whose state is being set.
        :param file_metadata: Submission file metadata to store
        :param state: A dictionary containing the file's state data to be logged.
        """
        file_path = Path(file_path)
        index = self._get_index(file_path)

        if file_metadata and not isinstance(file_metadata, SubmissionFileMetadata):
            file_metadata = SubmissionFileMetadata(**file_metadata)
        file_metadata = typing.cast(SubmissionFileMetadata, file_metadata)

        # Update state in memory
        self._file_states[index] = (file_metadata, state)

        # Persist state to JSON log file
        with open(self._file_path, "a", newline="") as fd:
            # Append the new state row to the log file
            json.dump(
                {
                    # index keys
                    **{k: v for k, v in zip(self._index.keys(), index, strict=True)},
                    # state
                    "metadata": file_metadata.model_dump(by_alias=True),
                    "state": state,
                },
                fd,
            )
            fd.write("\n")

    def num_entries(self) -> int:
        """
        Returns the number of entries in the file_states dictionary

        :return: An integer representing the number of entries in the file_states dictionary
        """
        return len(self._file_states)



"""Hash calculation utilities."""

import hashlib
import logging
from os import PathLike
from os.path import getsize
from pathlib import Path

from tqdm.auto import tqdm

from ..constants import TQDM_DEFAULTS

log = logging.getLogger(__name__)


def calculate_sha256(file_path: str | PathLike, chunk_size=2**16, progress=True) -> str:
    """
    Calculate the sha256 value of a file in chunks

    :param file_path: path to the file
    :param chunk_size: Chunk size in bytes
    :param progress: Print progress
    :return: calculated sha256 value of file_path
    """
    file_path = Path(file_path)
    total_size = getsize(file_path)
    sha256_hash = hashlib.sha256()
    with open(file_path, "rb") as f:
        if progress and (total_size > chunk_size):
            with tqdm(total=total_size, desc="SHA256  ", postfix=f"{file_path.name}", **TQDM_DEFAULTS) as pbar:  # type: ignore[call-overload]
                while chunk := f.read(chunk_size):
                    sha256_hash.update(chunk)
                    pbar.update(len(chunk))
        else:
            while chunk := f.read(chunk_size):
                sha256_hash.update(chunk)
    return sha256_hash.hexdigest()



"""Worker class for handling submission processing."""

from __future__ import annotations

import logging
import shutil
from os import PathLike
from pathlib import Path

from ..models.identifiers import IdentifiersModel
from ..models.s3 import S3Options
from ..validation import UserInterruptException
from .download import S3BotoDownloadWorker
from .submission import EncryptedSubmission, Submission, SubmissionValidationError

log = logging.getLogger(__name__)


class Worker:
    """Worker class for handling submission processing"""

    __log = log.getChild("Worker")

    def __init__(
        self,
        metadata_dir: str | PathLike,
        files_dir: str | PathLike,
        log_dir: str | PathLike,
        encrypted_files_dir: str | PathLike,
        threads: int = 1,
    ):
        """
        Initialize the worker object.

        :param metadata_dir: Path to the metadata directory
        :param files_dir: Path to the files directory
        :param log_dir: Path to the log directory
        :param encrypted_files_dir: Path to the encrypted files directory
        :param threads: Number of threads to use
        """
        self._threads = threads
        self.__log.debug("Threads: %s", self._threads)

        # metadata dir
        self.metadata_dir = Path(metadata_dir)
        self.__log.debug("Metadata directory: %s", self.metadata_dir)

        # files dir
        self.files_dir = Path(files_dir)
        self.__log.debug("Files directory: %s", self.files_dir)

        # encrypted files dir
        self.encrypted_files_dir = Path(encrypted_files_dir) if encrypted_files_dir is not None else Path()

        self.__log.info("Encrypted files directory: %s", self.encrypted_files_dir)

        # log dir
        self.log_dir = Path(log_dir)
        self.__log.info("Log directory: %s", self.log_dir)

        # create log dir if non-existent
        if not self.log_dir.is_dir():
            self.__log.debug("Creating log directory...")
            self.log_dir.mkdir(mode=0o770, parents=False, exist_ok=False)

        self.progress_file_checksum_validation = self.log_dir / "progress_checksum_validation.cjson"
        self.progress_file_sequencing_data_validation = self.log_dir / "progress_sequencing_data_validation.cjson"
        self.progress_file_encrypt = self.log_dir / "progress_encrypt.cjson"
        self.progress_file_decrypt = self.log_dir / "progress_decrypt.cjson"
        self.progress_file_upload = self.log_dir / "progress_upload.cjson"
        self.progress_file_download = self.log_dir / "progress_download.cjson"

    def parse_submission(self) -> Submission:
        """
        Reads the submission metadata and returns a Submission instance
        """
        submission = Submission(
            metadata_dir=self.metadata_dir,
            files_dir=self.files_dir,
        )
        return submission

    def parse_encrypted_submission(self) -> EncryptedSubmission:
        """
        Reads the submission metadata and returns an EncryptedSubmission instance
        """
        encrypted_submission = EncryptedSubmission(
            metadata_dir=self.metadata_dir, encrypted_files_dir=str(self.encrypted_files_dir), log_dir=self.log_dir
        )
        return encrypted_submission

    def validate(self, identifiers: IdentifiersModel, force=False, with_grz_check=True):
        """
        Validate this submission

        :param identifiers: IdentifiersModel containing GRZ and LE identifiers.
        :param force: Force validation of already validated files
        :param with_grz_check: If True, use the grz-check tool for validation.
        :raises SubmissionValidationError: if the validation fails
        """
        submission = self.parse_submission()

        self.__log.info("Starting metadata validation...")
        if errors := list(submission.metadata.validate(identifiers)):
            error_msg = "\n".join(["Metadata validation failed! Errors:", *errors])
            self.__log.error(error_msg)

            raise SubmissionValidationError(error_msg)
        else:
            self.__log.info("Metadata validation successful!")

        if force:
            # delete the log files if they exist
            self.progress_file_checksum_validation.unlink(missing_ok=True)
            self.progress_file_sequencing_data_validation.unlink(missing_ok=True)

        have_grz_check = shutil.which("grz-check") is not None
        if with_grz_check and have_grz_check:
            try:
                self.__log.info("Starting file validation with `grz-check`...")
                errors = list(
                    submission.validate_files_with_grz_check(
                        checksum_progress_file=self.progress_file_checksum_validation,
                        seq_data_progress_file=self.progress_file_sequencing_data_validation,
                        threads=self._threads,
                    )
                )
                if errors:
                    error_msg = "\n".join(["File validation failed! Errors:", *errors])
                    self.__log.error(error_msg)
                    raise SubmissionValidationError(error_msg)
                else:
                    self.__log.info("File validation successful!")
                return
            except UserInterruptException as e:
                error_msg = "Validation was cancelled by the user and is incomplete."
                self.__log.error(error_msg)
                raise SubmissionValidationError(error_msg) from e

        # Fallback validation
        self.__log.info("Starting checksum validation (fallback)...")
        if errors := list(
            submission._validate_checksums_fallback(progress_log_file=self.progress_file_checksum_validation)
        ):
            error_msg = "\n".join(["Checksum validation failed! Errors:", *errors])
            self.__log.error(error_msg)
            raise SubmissionValidationError(error_msg)
        else:
            self.__log.info("Checksum validation successful!")

        self.__log.info("Starting sequencing data validation (fallback)...")
        if errors := list(
            submission._validate_sequencing_data_fallback(
                progress_log_file=self.progress_file_sequencing_data_validation
            )
        ):
            error_msg = "\n".join(["Sequencing data validation failed! Errors:", *errors])
            self.__log.error(error_msg)
            raise SubmissionValidationError(error_msg)
        else:
            self.__log.info("Sequencing data validation successful!")

    def encrypt(
        self,
        recipient_public_key_path: str | PathLike,
        submitter_private_key_path: str | PathLike | None = None,
        force: bool = False,
    ) -> EncryptedSubmission:
        """
        Encrypt this submission with a public key using Crypt4Gh.
        :param recipient_public_key_path: Path to the public key file of the recipient.
        :param submitter_private_key_path: Path to the private key file of the submitter.
        :param force: Force encryption of already encrypted files
        :return: EncryptedSubmission instance
        """
        submission = self.parse_submission()

        if force:
            # delete the log file if it exists
            self.progress_file_encrypt.unlink(missing_ok=True)

        encrypted_submission = submission.encrypt(
            encrypted_files_dir=str(self.encrypted_files_dir),
            progress_log_file=self.progress_file_encrypt,
            recipient_public_key_path=recipient_public_key_path,
            submitter_private_key_path=submitter_private_key_path,
            force=force,
        )

        return encrypted_submission

    def decrypt(self, recipient_private_key_path: str | PathLike, force: bool = False) -> Submission:
        """
        Encrypt this submission with a public key using Crypt4Gh.
        :param recipient_private_key_path: Path to the private key file of the recipient.
        :param force: Force decryption of already decrypted files
        :return: EncryptedSubmission instance
        """
        encrypted_submission = self.parse_encrypted_submission()

        if force:
            # delete the log file if it exists
            self.progress_file_decrypt.unlink(missing_ok=True)

        submission = encrypted_submission.decrypt(
            files_dir=self.files_dir,
            progress_log_file=self.progress_file_decrypt,
            recipient_private_key_path=recipient_private_key_path,
        )

        return submission

    def upload(self, s3_options: S3Options) -> str:
        """
        Upload an encrypted submission and return the generated submission ID
        """
        from .upload import S3BotoUploadWorker

        upload_worker = S3BotoUploadWorker(
            s3_options, status_file_path=self.progress_file_upload, threads=self._threads
        )

        encrypted_submission = self.parse_encrypted_submission()

        upload_worker.upload(encrypted_submission)

        return encrypted_submission.submission_id

    def archive(self, s3_options: S3Options):
        """
        Archive an encrypted submission at a GRZ.
        """
        from .upload import S3BotoUploadWorker

        upload_worker = S3BotoUploadWorker(
            s3_options, status_file_path=self.progress_file_upload, threads=self._threads
        )

        encrypted_submission = self.parse_encrypted_submission()

        upload_worker.archive(encrypted_submission)

    def download(self, s3_options: S3Options, submission_id: str, force: bool = False):
        """
        Download an encrypted submission
        """
        if force:
            # delete the log file if it exists
            self.progress_file_download.unlink(missing_ok=True)

        download_worker = S3BotoDownloadWorker(
            s3_options, status_file_path=self.progress_file_download, threads=self._threads
        )

        self.__log.info("Preparing output directories...")
        download_worker.prepare_download(self.metadata_dir, self.encrypted_files_dir, self.log_dir)

        self.__log.info("Downloading metadata...")
        download_worker.download_metadata(submission_id, self.metadata_dir, metadata_file_name="metadata.json")

        self.__log.info("Downloading encrypted files...")
        download_worker.download(submission_id, EncryptedSubmission(self.metadata_dir, self.encrypted_files_dir))



"""Constants for logging configuration, JSON schema validation, and other settings."""

# This will be overridden by the package that imports it
PACKAGE_ROOT = "grz_common"

LOGGING_FORMAT = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
LOGGING_DATEFMT = "%Y-%m-%d %I:%M %p"
LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {
            "format": LOGGING_FORMAT,
            "datefmt": LOGGING_DATEFMT,
        },
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "WARNING",
            "propagate": False,
        },
        PACKAGE_ROOT: {"handlers": ["default"], "level": "INFO", "propagate": False},
        "__main__": {  # if __name__ == '__main__'
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
}

TQDM_SMOOTHING: float = 0.00001
TQDM_BAR_FORMAT = "{desc} ▕{bar:50}▏ {n_fmt:>10}/{total_fmt:<10} ({rate_fmt:>12}, ETA: {remaining:>6}) {postfix}"
TQDM_DEFAULTS = {
    "bar_format": TQDM_BAR_FORMAT,
    "unit": "iB",
    "unit_scale": True,
    "miniters": 1,
    "smoothing": TQDM_SMOOTHING,
    "colour": "cyan",
}

REDACTED_TAN = "0" * 64



"""
This module contains the type definitions for the progress logging states.
"""

from typing import TypedDict


class State(TypedDict, total=False):
    """
    Base state class for progress logging.
    """

    errors: list[str]


class ValidationState(State):
    """
    State for validation progress.
    """

    validation_passed: bool


class UploadState(State):
    """
    State for upload progress.
    """

    upload_successful: bool


class EncryptionState(State):
    """
    State for encryption progress.
    """

    encryption_successful: bool


class DecryptionState(State):
    """
    State for decryption progress.
    """

    decryption_successful: bool


class DownloadState(State):
    """
    State for download progress.
    """

    download_successful: bool



"""Validation package for grz-cli."""

import logging
import subprocess


class UserInterruptException(Exception):
    """Raised when an interrupt is triggered via ctrl+c."""

    pass


def run_grz_check(args: list[str]) -> subprocess.CompletedProcess:
    """
    Run `grz-check` with the given args.

    We catch KeyboardInterrupt here to allow the `grz-check` process to handle its own graceful shutdown
    and updating the progress logs from the report entries `grz-check` has generated so far.

    :param args: Arguments to pass to `grz-check`.
    :raises UserInterruptException: If interrupted via ctrl+c, `grz-check` exits with code 130.
    :raises subprocess.CalledProcessError: If `grz-check` fails for other reasons.
    """

    command = ["grz-check", *args]
    logging.info(f"Executing command: {' '.join(command)}")

    try:
        proc = subprocess.run(command, check=False)  # noqa: S603
    except KeyboardInterrupt:
        logging.warning("\nInterrupt received, allowing `grz-check` to shut down gracefully...")
        raise UserInterruptException from None

    if proc.returncode == 130:
        logging.warning("`grz-check` shut down gracefully due to an interrupt.")
        raise UserInterruptException
    elif proc.returncode != 0:
        logging.error(f"`grz-check` failed with a non-zero exit code: {proc.returncode}")
        raise subprocess.CalledProcessError(proc.returncode, command)
    return proc



"""Path utilities."""

import logging
import os
from os import PathLike

log = logging.getLogger(__name__)


def is_relative_subdirectory(relative_path: str | PathLike, root_directory: str | PathLike) -> bool:
    """
    Check if the target path is a subdirectory of the root path
    using os.path.commonpath() without checking the file system.

    :param relative_path: The target path.
    :param root_directory: The root directory.
    :return: True if relative_path is a subdirectory of root_directory, otherwise False.
    """
    # Convert both paths to absolute paths without resolving symlinks
    root_directory = os.path.abspath(root_directory)
    relative_path = os.path.abspath(relative_path)

    common_path = os.path.commonpath([root_directory, relative_path])

    # Check if the common path is equal to the root path
    return common_path == root_directory



from grz_pydantic_models.submission.metadata import GenomicDataCenterId, SubmitterId

from .base import IgnoringBaseModel, IgnoringBaseSettings


class IdentifiersModel(IgnoringBaseModel):
    grz: GenomicDataCenterId
    """Id of the GRZ."""

    le: SubmitterId
    """Id of the Leistungserbringer."""


class IdentifiersConfigModel(IgnoringBaseSettings):
    identifiers: IdentifiersModel



"""Utilities for handling crypt4gh keys, encryption and decryption"""

import io
import logging
import os
import typing
from functools import partial
from getpass import getpass
from os import PathLike
from os.path import getsize
from pathlib import Path

import crypt4gh.header
import crypt4gh.keys
import crypt4gh.lib
from nacl.public import PrivateKey
from tqdm.auto import tqdm

from ..constants import TQDM_DEFAULTS
from .io import TqdmIOWrapper

log = logging.getLogger(__name__)


class Crypt4GH:
    """Crypt4GH encryption/decryption utility class"""

    Key = tuple[int, bytes, bytes]

    VERSION = 1
    SEGMENT_SIZE = 65536
    FILE_EXTENSION = ".c4gh"

    @staticmethod
    def prepare_c4gh_keys(
        recipient_key_file_path: str | PathLike,
        sender_private_key: str | PathLike | None = None,
    ) -> tuple[Key]:
        """
        Prepare the key format that Crypt4GH needs. While it can contain multiple
         keys for multiple recipients, in our use case there is only a single recipient.

        :param recipient_key_file_path: path to the public key file of the recipient
        :param sender_private_key: path to the private key file of the sender.
            If None, will be generated randomly.
        """
        if sender_private_key is not None:
            sk = Crypt4GH.retrieve_private_key(sender_private_key)
        else:
            sk = bytes(PrivateKey.generate())
        keys = ((0, sk, crypt4gh.keys.get_public_key(recipient_key_file_path)),)
        return keys

    @staticmethod
    def encrypt_file(
        input_path: str | PathLike,
        output_path: str | PathLike,
        public_keys: tuple[Key],
    ):
        """
        Encrypt the file, properly handling the Crypt4GH header.

        :param public_keys:
        :param output_path:
        :param input_path:
        :return: tuple with md5 values for original file, encrypted file
        """
        # TODO: Progress bar?
        # TODO: store header in separate file?
        input_path = Path(input_path)
        output_path = Path(output_path)

        total_size = getsize(input_path)
        with (
            open(input_path, "rb") as in_fd,
            open(output_path, "wb") as out_fd,
            TqdmIOWrapper(
                typing.cast(io.RawIOBase, in_fd),
                tqdm(total=total_size, desc="ENCRYPT ", postfix=f"{input_path.name}", **TQDM_DEFAULTS),  # type: ignore[call-overload]
            ) as pbar_in_fd,
        ):
            crypt4gh.lib.encrypt(
                keys=public_keys,
                infile=pbar_in_fd,
                outfile=out_fd,
            )

    @staticmethod
    def retrieve_private_key(seckey_path) -> bytes:
        """
        Read Crypt4GH private key from specified path.
        :param seckey_path: path to the private key
        :return:
        """
        seckeypath = os.path.expanduser(seckey_path)
        if not os.path.exists(seckeypath):
            raise ValueError("Secret key not found")

        passphrase = os.getenv("C4GH_PASSPHRASE")
        if passphrase:
            passphrase_callback = lambda: passphrase
        else:
            passphrase_callback = partial(getpass, prompt=f"Passphrase for {seckey_path}: ")

        return crypt4gh.keys.get_private_key(seckeypath, passphrase_callback)

    @staticmethod
    def decrypt_file(input_path: Path, output_path: Path, private_key: bytes):
        """
        Decrypt a file using the provided private key
        :param input_path: Path to the encrypted file
        :param output_path: Path to the decrypted file
        :param private_key: The private key
        """
        total_size = getsize(input_path)
        file_name = input_path.name
        with (
            open(input_path, "rb") as in_fd,
            open(output_path, "wb") as out_fd,
            TqdmIOWrapper(
                typing.cast(io.RawIOBase, in_fd),
                tqdm(total=total_size, desc="DECRYPT ", postfix=f"{file_name}", **TQDM_DEFAULTS),  # type: ignore[call-overload]
            ) as pbar_in_fd,
        ):
            crypt4gh.lib.decrypt(
                keys=[(0, private_key, None)],  # list of (method, privkey, recipient_pubkey=None),
                infile=pbar_in_fd,
                outfile=out_fd,
            )



"""
Common click options for the CLI commands.
"""

from os import sched_getaffinity
from pathlib import Path

import click
import platformdirs

DEFAULT_CONFIG_PATH = Path(platformdirs.user_config_dir("grz-cli")) / "config.yaml"

# Aliases for path types for click options
# Naming convention: {DIR,FILE}_{Read,Write}_{Exists,Create}
DIR_R_E = click.Path(
    exists=True,
    file_okay=False,
    dir_okay=True,
    readable=True,
    writable=False,
    resolve_path=True,
)
DIR_RW_C = click.Path(
    exists=False,
    file_okay=False,
    dir_okay=True,
    readable=True,
    writable=True,
    resolve_path=True,
)
FILE_R_E = click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True)

submission_dir = click.option(
    "--submission-dir",
    metavar="PATH",
    type=DIR_R_E,
    required=True,
    help="Path to the submission directory containing 'metadata/', 'files/', 'encrypted_files/' and 'logs/' directories",
)

config_file = click.option(
    "--config-file",
    metavar="STRING",
    type=FILE_R_E,
    required=False,
    default=DEFAULT_CONFIG_PATH,
    help="Path to config file",
)

threads = click.option(
    "--threads",
    default=min(len(sched_getaffinity(0)), 4),
    type=int,
    show_default=True,
    help="Number of threads to use for parallel operations",
)

submission_id = click.option(
    "--submission-id",
    required=True,
    type=str,
    metavar="STRING",
    help="S3 submission ID",
)

output_dir = click.option(
    "--output-dir",
    metavar="PATH",
    type=DIR_RW_C,
    required=True,
    default=None,
    help="Path to the target submission output directory",
)

output_json = click.option("--json", "output_json", is_flag=True, help="Output JSON for machine-readability.")

show_details = click.option("--details", "show_details", is_flag=True, help="Show more detailed output.")

force = click.option("--force/--no-force", help="Overwrite files and ignore cached results (dangerous!)")



"""Validation of FASTQ files. Boils down to basic sanity checks such as line count and read length."""

import gzip
import logging
import os
import typing
from collections.abc import Generator
from contextlib import contextmanager
from gzip import GzipFile
from io import RawIOBase
from os import PathLike
from pathlib import Path
from typing import TextIO

from tqdm.auto import tqdm

from ..constants import TQDM_DEFAULTS
from ..utils.io import TqdmIOWrapper

log = logging.getLogger(__name__)


def is_gzipped(file_path: str | PathLike) -> bool:
    """
    Check if a file is gzipped based on its extension.

    :param file_path: Path to the file
    :return: True if the file is gzipped, False otherwise
    """
    return str(file_path).endswith(".gz")


@contextmanager
def open_fastq(file_path: str | PathLike, progress=True) -> Generator[TextIO, None, None]:
    """
    Open a FASTQ file, handling both regular and gzipped formats.

    :param file_path: Path to the FASTQ file
    :param progress: Whether to show a progress bar
    :return: A file object opened in the appropriate mode (gzipped or plain text)
    """
    handle: TqdmIOWrapper | GzipFile | typing.BinaryIO | None = None
    file_name = Path(file_path).name
    with open(file_path, "rb") as fd:
        # Open FASTQ
        total_size = os.stat(file_path).st_size
        if progress:
            handle = TqdmIOWrapper(
                typing.cast(RawIOBase, fd),
                tqdm(total=total_size, desc="FASTQ   ", postfix=f"{file_name}", **TQDM_DEFAULTS),  # type: ignore[call-overload]
            )
        else:
            handle = fd

        if is_gzipped(file_path):
            # decompress
            with gzip.open(typing.cast(RawIOBase, handle), "rb") as decompressed_fd:
                yield typing.cast(TextIO, decompressed_fd)
        else:
            yield typing.cast(TextIO, handle)


def calculate_fastq_stats(file_path, expected_read_length: int | None = None) -> tuple[int, int | None]:
    """
    Calculate line number and read lengths in FASTQ file.
    :param file_path: Path to the FASTQ file
    :param expected_read_length: Expected read length (None if not known)
    :return: tuple with the following values:
      - Number of lines in the file
      - Observed read length
    """
    with open_fastq(file_path) as f:
        read_length_warned = False
        for line_number, line in enumerate(f):
            if (line_number % 4) == 1:
                # Sequence lines are every 4th line starting from the 2nd
                # Check if the read length is consistent
                read_length = len(line.strip())
                if expected_read_length is None:
                    expected_read_length = read_length
                elif (not read_length_warned) and (read_length != expected_read_length):
                    # For the time being, read length mismatch is downgraded to warning
                    log.warning(
                        f"Read length mismatch at line {line_number + 1}: "
                        f"expected {expected_read_length}, found {read_length}. "
                        "This will be an error in the future."
                        "Further mismatches in the same file won't be reported."
                    )
                    read_length_warned = True

    return (
        line_number + 1,  # enumerate starts indexing at 0
        expected_read_length,
    )


def validate_fastq_file(
    fastq_file: str | PathLike, expected_read_length: int | None = None
) -> tuple[int, int, list[str]] | tuple[int, int | None, list[str]]:
    """
    Validates a FASTQ file.

    :param fastq_file: Path to the fastq file
    :param expected_read_length: Expected read length (None if not known)
    :return: Tuple with the following fields:
      - number of lines
      - set of observed read lengths
      - list of errors found
    """
    errors = []
    try:
        if expected_read_length is not None:
            log.debug("%s: expecting read length of %s", fastq_file, expected_read_length)

        # Calculate the number of lines and read lengths
        num_lines, read_length = calculate_fastq_stats(fastq_file, expected_read_length=expected_read_length)
    except ValueError as e:
        if "Read length mismatch" in str(e):
            log.warning(f"{fastq_file}: {e}")
        else:
            errors.append(f"{fastq_file}: {e}")
        return -1, -1, errors

    # Check if the number of lines in a FASTQ file is a multiple of 4.
    if num_lines % 4 != 0:
        errors.append(f"{fastq_file}: Number of lines is not a multiple of 4! Found {num_lines} lines.")
    else:
        log.debug("%s: %s lines", fastq_file, num_lines)

    return num_lines, read_length, errors


def validate_single_end_reads(fastq_file: str | PathLike, expected_read_length: int | None = None) -> Generator[str]:
    """
    Validate a single-end FASTQ file.

    :param fastq_file: Path to the FASTQ file
    :param expected_read_length: Expected read length (None if not known)
    :return: Generator of errors, if any.
    """
    num_lines, read_lengths, errors = validate_fastq_file(fastq_file, expected_read_length=expected_read_length)
    yield from errors


def validate_paired_end_reads(
    fastq_file1: str | PathLike,
    fastq_file2: str | PathLike,
    expected_read_length: int | None = None,
) -> Generator[str]:
    """
    Validate two paired-end FASTQ files.

    :param fastq_file1: Path to the first FASTQ file (Read 1)
    :param fastq_file2: Path to the second FASTQ file (Read 2)
    :param expected_read_length: Expected read length (None if not known)
    :return: Generator of errors, if any.
    """
    num_lines_file1, read_lengths_file1, errors_file1 = validate_fastq_file(
        fastq_file1, expected_read_length=expected_read_length
    )
    yield from errors_file1
    num_lines_file2, read_lengths_file2, errors_file2 = validate_fastq_file(
        fastq_file2, expected_read_length=expected_read_length
    )
    yield from errors_file2

    if num_lines_file1 != num_lines_file2:
        yield f"Paired-end files have different read counts: '{fastq_file1}' ({num_lines_file1}) and '{fastq_file2}' ({num_lines_file2})!"



from os import PathLike
from pathlib import Path
from typing import Annotated, Self

from pydantic import AfterValidator, BaseModel, ConfigDict
from pydantic.types import PathType
from pydantic_settings import BaseSettings, SettingsConfigDict

FilePath = Annotated[Path, AfterValidator(lambda v: v.expanduser()), PathType("file")]


class IgnoringBaseModel(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        validate_assignment=True,
        use_enum_values=True,
    )

    def to_yaml(self, fd):
        """Reads the configuration file and validates it against the schema."""
        import yaml

        yaml.dump(self.model_dump(mode="json", exclude_none=True, exclude_unset=True, exclude_defaults=True))

    @classmethod
    def from_path(cls, path: str | PathLike) -> Self:
        """Reads the configuration file and validates it against the schema."""
        import yaml

        with open(path, encoding="utf-8") as f:
            config = cls(**yaml.safe_load(f))  # noqa:

        return config


class IgnoringBaseSettings(BaseSettings):
    model_config = SettingsConfigDict(
        extra="ignore",
        validate_assignment=True,
        use_enum_values=True,
        env_nested_delimiter="__",
        env_prefix="grz_",
    )

    def to_yaml(self, fd):
        """Reads the configuration file and validates it against the schema."""
        import yaml

        yaml.dump(self.model_dump(mode="json", exclude_none=True, exclude_unset=True, exclude_defaults=True), fd)

    @classmethod
    def from_path(cls, path: str | PathLike) -> Self:
        """Reads the configuration file and validates it against the schema."""
        import yaml

        with open(path, encoding="utf-8") as f:
            config = cls(**yaml.safe_load(f))  # noqa:

        return config



"""Module for downloading encrypted submissions to local storage"""

from __future__ import annotations

import datetime
import enum
import itertools
import logging
import math
import re
from collections import OrderedDict
from operator import attrgetter, itemgetter
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING

import botocore.handlers
from boto3.s3.transfer import S3Transfer, TransferConfig  # type: ignore[import-untyped]
from grz_pydantic_models.submission.metadata.v1 import File as SubmissionFileMetadata
from pydantic import BaseModel
from tqdm.auto import tqdm

from ..constants import TQDM_DEFAULTS
from ..models.s3 import S3Options
from ..progress import DownloadState, FileProgressLogger
from ..transfer import init_s3_client

MULTIPART_THRESHOLD = 8 * 1024 * 1024  # 8MiB, boto3 default
MULTIPART_CHUNKSIZE = 8 * 1024 * 1024  # 8MiB, boto3 default
MULTIPART_MAX_CHUNKS = 1000  # CEPH S3 limit, AWS limit is 10000

if TYPE_CHECKING:
    from .submission import EncryptedSubmission

log = logging.getLogger(__name__)

# see discussion: https://github.com/boto/boto3/discussions/4251 to accept bucket names with ":" in the name
botocore.handlers.VALID_BUCKET = re.compile(r"^[:a-zA-Z0-9.\-_]{1,255}$")


class DownloadError(Exception):
    """Exception raised when an upload fails"""

    pass


class S3BotoDownloadWorker:
    """Implementation of a download worker using boto3 for S3"""

    __log = log.getChild("S3BotoDownloadWorker")

    def __init__(
        self,
        s3_options: S3Options,
        status_file_path: str | PathLike,
        threads: int = 1,
    ):
        """
        A download manager for S3 storage

        :param s3_options: The S3 configuration options
        :param status_file_path: The path to the status file
        :param threads: The number of concurrent download threads
        """
        super().__init__()

        self._status_file_path = Path(status_file_path)
        self._s3_options = s3_options
        self._threads = threads

        self._s3_client = init_s3_client(s3_options)

    def prepare_download(
        self,
        metadata_dir: Path,
        encrypted_files_dir: Path,
        log_dir: Path,
    ):
        """
        Prepare the download of an encrypted submission

        :param metadata_dir: Path to the metadata directory
        :param encrypted_files_dir: Path to the encrypted_files directory
        :param log_dir: Path to the logs directory
        """
        for dir_path in [metadata_dir, encrypted_files_dir, log_dir]:
            if not dir_path.exists():
                self.__log.debug("Creating directory: %s", dir_path)
                dir_path.mkdir(parents=False, exist_ok=False)
            else:
                self.__log.debug("Directory exists: %s", dir_path)

    def download_metadata(
        self,
        submission_id: str,
        metadata_dir: Path,
        metadata_file_name: str = "metadata.json",
    ):
        """
        Download the metadata.json

        :param submission_id: submission folder on S3 structure
        :param metadata_dir: Path of the metadir folder
        :param metadata_file_name: name of the metadata.json
        """
        metadata_key = str(Path(submission_id) / metadata_dir.name / metadata_file_name)
        metadata_file_path = metadata_dir / metadata_file_name

        self.__log.info("Downloading metadata file: '%s'", metadata_key)
        try:
            # Ensure the local target directory exists
            metadata_file_path.parent.mkdir(mode=0o770, parents=True, exist_ok=True)

            self._s3_client.download_file(self._s3_options.bucket, metadata_key, str(metadata_file_path))
            self.__log.info("Metadata download complete.")
        except botocore.exceptions.ClientError as e:
            if e.response.get("Error", {}).get("Code") == "404":
                error_msg = f"Metadata file '{metadata_key}' not found in S3 bucket '{self._s3_options.bucket}'."
                self.__log.error(error_msg)
                raise DownloadError(error_msg) from e
            raise e
        except Exception as e:
            self.__log.error("Download failed for metadata '%s'", metadata_key)
            raise e

    def _download_with_progress(self, local_file_path: str, s3_object_id: str):
        """
        Download a single file from S3 to local storage.

        :param local_file_path: Path to the local target file.
        :param s3_object_id: The S3 object key to download.
        """
        s3_object_meta = self._s3_client.head_object(Bucket=self._s3_options.bucket, Key=s3_object_id)
        filesize = s3_object_meta["ContentLength"]

        chunksize = (
            math.ceil(filesize / MULTIPART_MAX_CHUNKS)
            if filesize / MULTIPART_CHUNKSIZE > MULTIPART_MAX_CHUNKS
            else MULTIPART_CHUNKSIZE
        )
        self.__log.debug(
            f"Using a chunksize of: {chunksize / 1024**2}MiB, results in {math.ceil(filesize / chunksize)} chunks"
        )

        config = TransferConfig(
            multipart_threshold=MULTIPART_THRESHOLD,
            multipart_chunksize=chunksize,
            max_concurrency=self._threads,
        )

        transfer = S3Transfer(self._s3_client, config)  # type: ignore[arg-type]
        with tqdm(total=filesize, postfix=f"{s3_object_id}", **TQDM_DEFAULTS) as progress_bar:  # type: ignore[call-overload]
            transfer.download_file(
                self._s3_options.bucket,
                s3_object_id,
                local_file_path,
                callback=lambda bytes_transferred: progress_bar.update(bytes_transferred),
            )

    def download_file(
        self,
        local_file_path: Path,
        s3_object_id: str,
        progress_logger: FileProgressLogger[DownloadState],
        file_metadata: SubmissionFileMetadata,
    ):
        """
        Download a single file from S3 to the specified local_file_path.

        :param local_file_path: Path to the local file.
        :param s3_object_id: S3 key of the file to download.
        :param progress_logger: The progress logger instance.
        :param file_metadata: The metadata for the file.
        """
        try:
            local_file_path.parent.mkdir(mode=0o770, parents=True, exist_ok=True)

            self._download_with_progress(str(local_file_path), s3_object_id)

            self.__log.info(f"Download complete for {str(local_file_path)}.")
            progress_logger.set_state(local_file_path, file_metadata, state=DownloadState(download_successful=True))

        except botocore.exceptions.ClientError as e:
            if e.response.get("Error", {}).get("Code") == "404":
                error_msg = f"File '{s3_object_id}' not found in S3 bucket '{self._s3_options.bucket}'."
                exc = DownloadError(error_msg)
            else:
                error_msg = f"S3 client error for '{s3_object_id}': {e}"
                exc = e  # type: ignore[assignment]
            self.__log.error(error_msg)
            progress_logger.set_state(
                local_file_path, file_metadata, state=DownloadState(download_successful=False, errors=[str(exc)])
            )
            raise exc from e
        except Exception as e:
            self.__log.error("Download failed for '%s': %s", str(local_file_path), e)
            progress_logger.set_state(
                local_file_path, file_metadata, state=DownloadState(download_successful=False, errors=[str(e)])
            )
            raise e

    def download(self, submission_id: str, encrypted_submission: EncryptedSubmission):
        """
        Download an encrypted submission.

        This method iterates through the files listed in the submission's metadata,
        constructs their S3 object keys, and downloads them.

        :param submission_id: The ID of the submission, used as a prefix in S3.
        :param encrypted_submission: The encrypted submission to download.
        """
        progress_logger = FileProgressLogger[DownloadState](self._status_file_path)

        for local_file_path, file_metadata in encrypted_submission.encrypted_files.items():
            relative_encrypted_path = file_metadata.encrypted_file_path()
            file_key = f"{submission_id}/files/{relative_encrypted_path}"

            logged_state = progress_logger.get_state(local_file_path, file_metadata)
            if logged_state and logged_state.get("download_successful"):
                self.__log.info(
                    "File '%s' already downloaded (at '%s'), skipping.",
                    file_key,
                    str(local_file_path),
                )
                continue

            self.__log.info("Downloading file: '%s' -> '%s'", file_key, str(local_file_path))
            self.download_file(local_file_path, file_key, progress_logger, file_metadata)


class InboxSubmissionState(enum.StrEnum):
    INCOMPLETE = "incomplete"
    COMPLETE = "complete"
    CLEANING = "cleaning"
    CLEANED = "cleaned"
    ERROR = "error"


class InboxSubmissionSummary(BaseModel):
    """A summary of the state of a submission in an inbox"""

    submission_id: str
    state: InboxSubmissionState
    oldest_upload: datetime.datetime
    newest_upload: datetime.datetime


def query_submissions(s3_options: S3Options, show_cleaned: bool) -> list[InboxSubmissionSummary]:
    """Queries the state of all submissions in the configured bucket."""
    s3_client = init_s3_client(s3_options)
    paginator = s3_client.get_paginator("list_objects_v2")

    objects = itertools.chain.from_iterable(
        page["Contents"] for page in paginator.paginate(Bucket=s3_options.bucket) if "Contents" in page
    )
    objects_sorted = sorted(objects, key=itemgetter("Key"))
    submission2objects = {
        key: tuple(group) for key, group in itertools.groupby(objects_sorted, key=lambda o: o["Key"].split("/")[0])
    }

    submissions = []
    for submission_id, submission_objects in submission2objects.items():
        submission_objects_sorted = OrderedDict(
            (o["Key"], o) for o in sorted(submission_objects, key=itemgetter("LastModified"))
        )
        oldest_object = submission_objects_sorted[next(iter(submission_objects_sorted))]
        newest_object = submission_objects_sorted[next(reversed(submission_objects_sorted))]

        cleaning_key = f"{submission_id}/cleaning"
        cleaned_key = f"{submission_id}/cleaned"
        if (cleaning_key in submission_objects_sorted) and (cleaned_key in submission_objects_sorted):
            log.warning("Submission '{submission_id}' is in an incomplete cleaned state!")
            state = InboxSubmissionState.ERROR
        elif cleaning_key in submission_objects_sorted:
            state = InboxSubmissionState.CLEANING
        elif cleaned_key in submission_objects_sorted:
            state = InboxSubmissionState.CLEANED
        else:
            state = (
                InboxSubmissionState.COMPLETE
                if f"{submission_id}/metadata/metadata.json" in submission_objects_sorted
                else InboxSubmissionState.INCOMPLETE
            )

        submission = InboxSubmissionSummary(
            submission_id=submission_id,
            state=state,
            oldest_upload=oldest_object["LastModified"],
            newest_upload=newest_object["LastModified"],
        )

        if state in {InboxSubmissionState.CLEANING, InboxSubmissionState.CLEANED} and (not show_cleaned):
            # skip listing cleaning/cleaned submissions unless show_cleaned is true
            continue

        submissions.append(submission)

    return sorted(submissions, key=attrgetter("oldest_upload"))



"""I/O utilities."""

import io
import json
import logging
from typing import TextIO

log = logging.getLogger(__name__)


def read_multiple_json(input_file: TextIO):
    """
    Read multiple JSON objects from a text stream.
    :param input_file:
    """
    for line in input_file:
        if line.strip():
            yield json.loads(line)
        else:
            raise ValueError("Encountered blank line while reading jsonl, invalid.")


class TqdmIOWrapper(io.RawIOBase):
    """
    Wrapper to record reads and writes in a tqdm progress bar.

    Example:
        very_long_input = io.StringIO("0123456789abcdef" * 100000)
        total_size = len(very_long_input.getvalue())

        batch_size = 10 ** 4  # 10kb
        with open("/dev/null", "w") as fd:
            with TqdmIOWrapper(fd, tqdm(
                    total=total_size,
                    desc="Printing: ",
                    unit="B",
                    unit_scale=True,
                    # unit_divisor=1024,  # make use of standard units e.g. KB, MB, etc.
                    miniters=1,
            )) as pbar_fd:
                while chunk := very_long_input.read(batch_size):
                    pbar_fd.write(chunk)

                    time.sleep(0.1)
    """

    def __init__(self, io_buf: io.RawIOBase, progress_bar):
        """

        :param io_buf: the buffer to wrap
        :param progress_bar: tqdm progress bar
        """
        self.io_buf = io_buf
        self.callback = progress_bar.update

    def write(self, data):
        """Write data to the buffer and update the progress bar"""
        nbytes_written = self.io_buf.write(data)
        if nbytes_written:
            self.callback(nbytes_written)
        return nbytes_written

    def read(self, size=-1) -> bytes | None:
        """Read data from the buffer and update the progress bar"""
        data = self.io_buf.read(size)
        if data:
            self.callback(len(data))

        return data

    def readinto(self, buffer, /):
        """Read data into a buffer and update the progress bar"""
        nbytes_written = self.io_buf.readinto(buffer)
        if nbytes_written:
            self.callback(nbytes_written)

        return nbytes_written

    def flush(self):
        """Flush the buffer"""
        # Ensure all data is flushed to the underlying binary IO object
        self.io_buf.flush()

    def close(self):
        """Close the buffer"""
        # Close the underlying binary IO object
        self.io_buf.close()



from typing import Self

from grz_common.models.base import FilePath, IgnoringBaseSettings
from pydantic import field_validator, model_validator


class KeyModel(IgnoringBaseSettings):
    grz_public_key: str | None = None
    """
    The public key of the recipient (the associated GRZ).
    """

    grz_public_key_path: FilePath | None = None
    """
    Path to the crypt4gh public key of the recipient (the associated GRZ).
    """

    grz_private_key_path: FilePath | None = None
    """
    Path to the crypt4gh private key of the recipient (optional).
    """

    submitter_private_key_path: FilePath | None = None
    """
    Path to the submitter's private key (optional).
    """

    @field_validator("grz_public_key")
    @classmethod
    def check_grz_public_key(cls, v):
        if v is not None and "BEGIN CRYPT4GH PUBLIC KEY" not in v and "END CRYPT4GH PUBLIC KEY" not in v:
            raise ValueError("Invalid public key format")
        return v

    @model_validator(mode="after")
    def validate_grz_public_key(self) -> Self:
        if self.grz_public_key is None and self.grz_public_key_path is None:
            raise ValueError("Either grz_public_key or grz_public_key_path must be set.")
        if self.grz_public_key is not None and self.grz_public_key_path is not None:
            raise ValueError("Only one of grz_public_key or grz_public_key_path must be set.")
        return self


class KeyConfigModel(IgnoringBaseSettings):
    keys: KeyModel



"""
Module: logging_setup

This module provides functions for setting up logging configuration.
"""

from __future__ import annotations

import logging
from os import PathLike
from pathlib import Path

from .constants import LOGGING_DATEFMT, LOGGING_FORMAT, PACKAGE_ROOT

log = logging.getLogger(__name__)


def add_filelogger(
    file_path: str | PathLike | None = None,
    level: str = "INFO",
    logger_name: str = PACKAGE_ROOT,
) -> None:
    """
    Add file logging for the specified package.

    This function configures a file logger to capture log messages
    for the package specified by logger_name. If no file path
    is provided, a default log file will be created in the user's
    home directory.

    :param file_path: Optional; the path to the log file. If None,
                      a default path will be used.
    :param level: Optional; the logging level. Default is 'INFO'.
                  Must be a valid logging level name (e.g., 'DEBUG', 'INFO').
    :param logger_name: Optional; the name of the logger to add the file handler to.
                        Default is PACKAGE_ROOT.
    """
    logger = logging.getLogger(logger_name)

    if file_path is None:
        default_log_dir = Path.home() / "logs"
        default_log_dir.mkdir(parents=True, exist_ok=True)
        file_path = default_log_dir / f"{logger_name}.log"
        log.warning("No log file path provided, using default: %s", file_path)
    else:
        file_path = Path(file_path)

    try:
        fh = logging.FileHandler(file_path)
        fh.setLevel(level.upper())
        fh.setFormatter(logging.Formatter(LOGGING_FORMAT, LOGGING_DATEFMT))
        logger.addHandler(fh)
        log.info(
            "File logger added for %s at %s with level %s.",
            logger.name,
            file_path,
            level.upper(),
        )
    except Exception as e:
        log.error("Failed to add file logger: %s", e)



"""Module for uploading encrypted submissions to a remote storage"""

from __future__ import annotations

import abc
import json
import logging
import math
import re
import shutil
from os import PathLike
from os.path import getsize
from pathlib import Path
from typing import TYPE_CHECKING, override

import botocore.handlers
from boto3.s3.transfer import S3Transfer, TransferConfig  # type: ignore[import-untyped]
from grz_common.constants import REDACTED_TAN
from tqdm.auto import tqdm

from ..constants import TQDM_DEFAULTS
from ..models.s3 import S3Options
from ..progress import FileProgressLogger, UploadState
from ..transfer import init_s3_client

MULTIPART_THRESHOLD = 8 * 1024**2  # 8MiB, boto3 default, largely irrelevant
MULTIPART_MAX_CHUNKS = 1000  # CEPH S3 limit, AWS limit is 10000

if TYPE_CHECKING:
    from .submission import EncryptedSubmission

log = logging.getLogger(__name__)

# see discussion: https://github.com/boto/boto3/discussions/4251 for acception bucketnames with : in the name
botocore.handlers.VALID_BUCKET = re.compile(r"^[:a-zA-Z0-9.\-_]{1,255}$")  # type: ignore[import-untyped]


class UploadError(Exception):
    """Exception raised when an upload fails"""

    pass


class UploadWorker(metaclass=abc.ABCMeta):
    """Worker baseclass for uploading encrypted submissions"""

    @abc.abstractmethod
    def upload(self, encrypted_submission: EncryptedSubmission):
        """
        Upload an encrypted submission to a GRZ inbox

        :param encrypted_submission: The encrypted submission to upload
        :raises UploadError: when the upload failed
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def upload_file(self, local_file_path: str | PathLike, s3_object_id: str):
        """
        Upload a single file to the specified object ID
        :param local_file_path: Path to the file to upload
        :param s3_object_id: Remote S3 object ID under which the file should be stored
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def archive(self, encrypted_submission: EncryptedSubmission):
        """
        Archive an encrypted submission within a GRZ

        :param encrypted_submission: The encrypted submission to archive
        :raises UploadError: when archival failed
        """
        raise NotImplementedError()


class S3BotoUploadWorker(UploadWorker):
    """Implementation of an upload worker using boto3 for S3"""

    __log = log.getChild("S3BotoUploadWorker")

    def __init__(
        self,
        s3_options: S3Options,
        status_file_path: str | PathLike,
        threads: int = 1,
    ):
        """
        An upload manager for S3 storage

        :param config: instance of `ConfigModel`
        :param status_file_path: file for storing upload state. Can be used for e.g. resumable uploads.
        """
        super().__init__()

        self._status_file_path = Path(status_file_path)
        self._s3_options = s3_options
        self._threads = threads

        self._s3_client = init_s3_client(s3_options)

    @override
    def upload_file(self, local_file_path: str | PathLike, s3_object_id: str):
        """
        Upload a single file to the specified object ID
        :param local_file_path: Path to the file to upload
        :param s3_object_id: Remote S3 object ID under which the file should be stored
        """
        self.__log.info(f"Uploading {local_file_path} to {s3_object_id}...")

        filesize = getsize(local_file_path)
        multipart_chunksize = self._s3_options.multipart_chunksize

        chunksize = (
            math.ceil(filesize / MULTIPART_MAX_CHUNKS)
            if filesize / multipart_chunksize > MULTIPART_MAX_CHUNKS
            else multipart_chunksize
        )
        self.__log.debug(
            f"Using a chunksize of: {chunksize / 1024**2}MiB, results in {math.ceil(filesize / chunksize)} chunk(s)"
        )

        config = TransferConfig(
            multipart_threshold=MULTIPART_THRESHOLD,
            multipart_chunksize=chunksize,
            max_concurrency=self._threads,
            use_threads=self._threads > 1,
        )

        transfer = S3Transfer(self._s3_client, config)  # type: ignore[arg-type]
        progress_bar = tqdm(total=filesize, desc="UPLOAD  ", **TQDM_DEFAULTS, postfix=f"{s3_object_id}")  # type: ignore[call-overload]
        transfer.upload_file(
            str(local_file_path),
            self._s3_options.bucket,
            s3_object_id,
            callback=lambda bytes_transferred: progress_bar.update(bytes_transferred),
        )

    def _remote_id_exists(self, s3_object_id: str) -> bool:
        """
        Determine if a remote ID already exists
        :param s3_object_id: Remote S3 object ID under which the file should be stored
        """
        exists = True
        try:
            self._s3_client.head_object(Bucket=self._s3_options.bucket, Key=s3_object_id)
        except self._s3_client.exceptions.NoSuchKey:
            exists = False
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] in {"403", "404"}:
                # backend can return forbidden instead if user has no ListBucket permission
                exists = False
            else:
                raise error

        return exists

    def _upload_logged_files(self, encrypted_submission, progress_logger, files_to_upload):
        for file_path in files_to_upload:
            if not Path(file_path).exists():
                raise UploadError(f"File {file_path} does not exist")

        for file_path, file_metadata in encrypted_submission.encrypted_files.items():
            logged_state = progress_logger.get_state(file_path, file_metadata)
            self.__log.debug("state for %s: %s", file_path, logged_state)

            s3_object_id = files_to_upload[file_path]

            if (logged_state is None) or not logged_state.get("upload_successful", False):
                self.__log.info(
                    "Uploading file: '%s' -> '%s'",
                    str(file_path),
                    str(s3_object_id),
                )

                try:
                    self.upload_file(file_path, s3_object_id)

                    self.__log.info(f"Upload complete for {str(file_path)}. ")
                    progress_logger.set_state(
                        file_path,
                        file_metadata,
                        state=UploadState(upload_successful=True),
                    )
                except Exception as e:
                    self.__log.error("Upload failed for '%s'", str(file_path))

                    progress_logger.set_state(
                        file_path,
                        file_metadata,
                        state=UploadState(upload_successful=False, errors=[str(e)]),
                    )

                    raise e
            else:
                self.__log.info(
                    "File '%s' already uploaded (at '%s')",
                    str(file_path),
                    str(s3_object_id),
                )

    def _upload_metadata(self, metadata_file_path, metadata_s3_object_id):
        # upload metadata unconditionally
        try:
            self.upload_file(metadata_file_path, metadata_s3_object_id)
            self.__log.info(f"Upload complete for {str(metadata_file_path)}. ")
        except Exception as e:
            self.__log.error("Upload failed for '%s'", str(metadata_file_path))
            raise e

    @override
    def upload(self, encrypted_submission: EncryptedSubmission):
        """
        Upload an encrypted submission
        :param encrypted_submission: The encrypted submission to upload
        """
        progress_logger = FileProgressLogger[UploadState](self._status_file_path)
        metadata_file_path, metadata_s3_object_id = encrypted_submission.get_metadata_file_path_and_object_id()

        if self._remote_id_exists(metadata_s3_object_id):
            raise UploadError("Submission already uploaded. Corrections, additions, and followups require a new tanG.")

        files_to_upload = encrypted_submission.get_encrypted_files_and_object_id()
        files_to_upload[metadata_file_path] = metadata_s3_object_id

        self._upload_logged_files(encrypted_submission, progress_logger, files_to_upload)

        self._upload_metadata(metadata_file_path, metadata_s3_object_id)

    @override
    def archive(self, encrypted_submission: EncryptedSubmission):
        """
        Archive an encrypted submission
        :param encrypted_submission: The encrypted submission to upload
        """
        progress_logger = FileProgressLogger[UploadState](self._status_file_path)
        metadata_file_path, metadata_s3_object_id = encrypted_submission.get_metadata_file_path_and_object_id()

        if self._remote_id_exists(metadata_s3_object_id):
            raise UploadError("Submission already archived.")

        files_to_upload = encrypted_submission.get_encrypted_files_and_object_id()
        files_to_upload[metadata_file_path] = metadata_s3_object_id

        log_file_to_object_id = encrypted_submission.get_log_files_and_object_id()
        overlapping_files = files_to_upload.keys() & log_file_to_object_id.keys()
        if overlapping_files:
            raise ValueError(
                f"Conflict in files specified for archive: {overlapping_files}. This is a bug. Please report this to the developers."
            )
        files_to_upload.update(log_file_to_object_id)

        self._upload_logged_files(encrypted_submission, progress_logger, files_to_upload)

        # do not track upload state for logs, instead just reupload in case of a failure
        for file_path, s3_object_id in encrypted_submission.get_log_files_and_object_id().items():
            try:
                self.upload_file(file_path, s3_object_id)
                self.__log.info(f"Upload complete for {str(file_path)}.")
            except Exception as e:
                self.__log.error("Upload failed for '%s'", str(file_path))
                raise e

        # make a back up copy of metadata before editing
        shutil.copy(metadata_file_path, metadata_file_path.with_suffix(".orig.json"))

        with open(metadata_file_path, mode="r+") as metadata_file:
            metadata = json.load(metadata_file)
            # redact tanG as all zeros
            metadata["submission"]["tanG"] = REDACTED_TAN

            # redact local case ID
            metadata["submission"]["localCaseId"] = ""

            for donor in metadata["donors"]:
                if donor["relation"] == "index":
                    # redact index donorPseudonym (which can be the tanG)
                    donor["donorPseudonym"] = "index"

            metadata_file.seek(0)
            json.dump(metadata, metadata_file, indent=2)
            metadata_file.truncate()

        self._upload_metadata(metadata_file_path, metadata_s3_object_id)

    def _check_for_completed_submission(self, s3_object_id: str) -> bool:
        try:
            return s3_object_id in self._list_keys(self._s3_options.bucket, prefix=str(Path(s3_object_id).parent))
        except Exception as e:
            self.__log.warning(
                "Exception occured during check for completed submission; assuming submission is incomplete.",
                exc_info=e,
            )
            return False

    # https://stackoverflow.com/a/54014862
    def _list_keys(self, bucket_name, prefix="/", delimiter="/", start_after=""):
        s3_paginator = self._s3_client.get_paginator("list_objects_v2")
        prefix = prefix.lstrip(delimiter)
        start_after = (start_after or prefix) if prefix.endswith(delimiter) else start_after
        for page in s3_paginator.paginate(Bucket=bucket_name, Prefix=prefix, StartAfter=start_after):
            for content in page.get("Contents", ()):
                yield content["Key"]



"""Progress tracking and logging functionality."""

from .progress_logging import FileProgressLogger
from .states import (
    DecryptionState,
    DownloadState,
    EncryptionState,
    State,
    UploadState,
    ValidationState,
)

__all__ = [
    "DecryptionState",
    "DownloadState",
    "EncryptionState",
    "FileProgressLogger",
    "State",
    "UploadState",
    "ValidationState",
]



"""Classes for parsing and validating submission metadata and files."""

from __future__ import annotations

import json
import logging
import subprocess
import typing
from collections.abc import Generator
from itertools import groupby
from os import PathLike
from pathlib import Path

from grz_pydantic_models.submission.metadata import get_accepted_versions
from grz_pydantic_models.submission.metadata.v1 import (
    ChecksumType,
    File,
    FileType,
    GrzSubmissionMetadata,
    ReadOrder,
    SequenceData,
    SequencingLayout,
)
from grz_pydantic_models.submission.metadata.v1 import File as SubmissionFileMetadata
from pydantic import ValidationError

from ..models.identifiers import IdentifiersModel
from ..progress import DecryptionState, EncryptionState, FileProgressLogger, ValidationState
from ..utils.checksums import calculate_sha256
from ..utils.crypt import Crypt4GH
from ..validation import UserInterruptException, run_grz_check
from ..validation.bam import validate_bam
from ..validation.fastq import validate_paired_end_reads, validate_single_end_reads

log = logging.getLogger(__name__)

S3_MAX_KEY_LENGTH = 1024
# length of uploaded prefix before LE-specified filepath
# e.g. len("123456789_2025-04-01_a70eb6ce/files/") == 36
UPLOADED_FILE_PREFIX_LENGTH = 36


class SubmissionMetadata:
    """Class for reading and validating submission metadata"""

    __log = log.getChild("SubmissionMetadata")

    def __init__(self, metadata_file: Path):
        """
        Load, parse and validate the metadata file.

        :param metadata_file: path to the metadata.json file
        :raises json.JSONDecodeError: if failed to read the metadata.json file
        :raises jsonschema.exceptions.ValidationError: if metadata does not match expected schema
        """
        self.file_path = metadata_file
        self.content = self._read_metadata(self.file_path)
        self._checksum = calculate_sha256(self.file_path, progress=False)

        self._files: dict | None = None

    @classmethod
    def _read_metadata(cls, file_path: Path) -> GrzSubmissionMetadata:
        """
        Load and parse the metadata file in JSON format.

        :param file_path: Path to the metadata JSON file
        :return: Parsed metadata as a dictionary
        :raises json.JSONDecodeError: if failed to read the metadata.json file
        """
        try:
            with open(file_path, encoding="utf-8") as jsonfile:
                metadata = json.load(jsonfile)
                try:
                    metadata_model = GrzSubmissionMetadata(**metadata)
                except ValidationError as ve:
                    cls.__log.error("Invalid metadata format in metadata file: %s", file_path)
                    raise SystemExit(ve) from ve
                return metadata_model
        except json.JSONDecodeError as e:
            cls.__log.error("Invalid JSON format in metadata file: %s", file_path)
            raise e

    @property
    def transaction_id(self) -> str:
        """
        The index case ID of this submission
        """
        return self.content.submission.tan_g

    @property
    def files(self) -> dict[Path, SubmissionFileMetadata]:
        """
        The files linked in the metadata.

        :return: Dictionary of `file_path` -> `SubmissionFileMetadata` pairs.
            Each `file_path` refers to the relative file path from the metadata.
        """
        if self._files is not None:
            return self._files

        submission_files = {}
        for donor in self.content.donors:
            for lab_data in donor.lab_data:
                if not lab_data.sequence_data:
                    continue
                for file_data in lab_data.sequence_data.files:
                    file_path = Path(file_data.file_path)
                    submission_files[file_path] = file_data

        self._files = submission_files
        return self._files

    def validate(self, identifiers: IdentifiersModel) -> Generator[str]:  # noqa: C901, PLR0912
        """
        Validates this submission's metadata (content).

        :return: Generator of errors
        """
        metadata_schema_version = self.content.get_schema_version()
        accepted_versions = get_accepted_versions()
        if metadata_schema_version not in accepted_versions:
            yield f"Metadata schema version {metadata_schema_version} is outdated. Currently accepting the following versions: {', '.join(accepted_versions)}"

        expected_grz_id, expected_le_id = identifiers.grz, identifiers.le
        if (submitted_grz_id := self.content.submission.genomic_data_center_id) != expected_grz_id:
            yield (
                f"Genomic data center identifier specified in the metadata.json ({submitted_grz_id}) "
                f"does not match genomic data center identifier in config ({expected_grz_id})"
            )

        if (submitted_le_id := self.content.submission.submitter_id) != expected_le_id:
            yield (
                f"Submitter (LE) identifier specified in the metadata.json ({submitted_le_id}) "
                f"does not match submitter (LE) identifier in config ({expected_le_id})"
            )

        submission_files: dict[str | PathLike, SubmissionFileMetadata] = {}
        for donor in self.content.donors:
            for lab_data in donor.lab_data:
                if not lab_data.sequence_data:
                    log.info(f"Skipping validation of empty sequence data for donor {donor}.")
                    continue
                for file_data in lab_data.sequence_data.files:
                    # check if file is already registered
                    file_path = Path(file_data.file_path)
                    if len(str(file_path)) > (S3_MAX_KEY_LENGTH - UPLOADED_FILE_PREFIX_LENGTH):
                        yield f"{file_data.file_path}: File path is too long for the inbox!"
                    if other_metadata := submission_files.get(file_path):
                        # check if metadata matches
                        if file_data != other_metadata:
                            yield f"{file_data.file_path}: Different metadata for the same path observed!"

                        # check if FASTQ data was already linked in another submission
                        if file_data.file_type == "fastq":
                            yield f"{file_data.file_path}: FASTQ file already linked in another submission!"
                        if file_data.file_type == "bam":
                            yield f"{file_data.file_path}: BAM file already linked in another submission!"
                    else:
                        submission_files[file_path] = file_data

    @property
    def checksum(self) -> str:
        """
        Checksum of the metadata file
        """
        return self._checksum


class Submission:
    """Class for handling submission data"""

    __log = log.getChild("Submission")

    def __init__(self, metadata_dir: str | PathLike, files_dir: str | PathLike):
        """
        Initialize the submission object.

        :param metadata_dir: Path to the metadata directory
        :param files_dir: Path to the files directory
        """
        self.metadata_dir = Path(metadata_dir)
        self.files_dir = Path(files_dir)

        self.metadata = SubmissionMetadata(self.metadata_dir / "metadata.json")

    @property
    def files(self) -> dict[Path, SubmissionFileMetadata]:
        """
        The files liked in the metadata.

        :return: Dictionary of `local_file_path` -> `SubmissionFileMetadata` pairs.
        """
        retval = {}
        for file_path, file_metadata in self.metadata.files.items():
            local_file_path = self.files_dir / file_path

            retval[local_file_path] = file_metadata

        return retval

    def validate_files_with_grz_check(  # noqa: C901, PLR0915, PLR0912
        self, checksum_progress_file: str | PathLike, seq_data_progress_file: str | PathLike, threads: int | None
    ) -> Generator[str, None, None]:
        """
        Validates submission files using `grz-check` and populates both progress logs.
        """
        checksum_progress_logger = FileProgressLogger[ValidationState](log_file_path=checksum_progress_file)
        checksum_progress_logger.cleanup(keep=[(fp, fm) for fp, fm in self.files.items()])

        seq_data_progress_logger = FileProgressLogger[ValidationState](log_file_path=seq_data_progress_file)
        seq_data_progress_logger.cleanup(keep=[(fp, fm) for fp, fm in self.files.items()])

        grz_check_args = []
        checked_files = set()

        def should_check_file(file_path: Path, file_metadata: File) -> bool:
            # Check against both logs. If either is missing a "pass", re-check.
            checksum_state = checksum_progress_logger.get_state(file_path, file_metadata)
            seq_data_state = seq_data_progress_logger.get_state(file_path, file_metadata)
            is_seq_file = file_metadata.file_type in ("fastq", "bam")

            checksum_passed = checksum_state and checksum_state.get("validation_passed")
            seq_data_passed = seq_data_state and seq_data_state.get("validation_passed")

            if is_seq_file:
                return not (checksum_passed and seq_data_passed)
            return not checksum_passed

        # Collect args for FASTQ and BAM from sequencing data
        for donor in self.metadata.content.donors:
            for lab_data in donor.lab_data:
                if not lab_data.sequence_data:
                    continue

                sequence_data = lab_data.sequence_data
                fastq_files = [f for f in sequence_data.files if f.file_type == FileType.fastq]
                bam_files = [f for f in sequence_data.files if f.file_type == FileType.bam]

                if lab_data.sequencing_layout == SequencingLayout.paired_end:
                    key = lambda f: (f.flowcell_id, f.lane_id)
                    fastq_files.sort(key=key)
                    for _key, group in groupby(fastq_files, key):
                        files = list(group)
                        fastq_r1_files = [f for f in files if f.read_order == ReadOrder.r1]
                        fastq_r2_files = [f for f in files if f.read_order == ReadOrder.r2]

                        for r1_meta, r2_meta in zip(fastq_r1_files, fastq_r2_files, strict=True):
                            r1_path = self.files_dir / r1_meta.file_path
                            r2_path = self.files_dir / r2_meta.file_path
                            if should_check_file(r1_path, r1_meta) or should_check_file(r2_path, r2_meta):
                                r1_read_len = -1  # disable read length check for now; r1_meta.read_length
                                r2_read_len = -1  # disable read length check for now; r2_meta.read_length
                                grz_check_args.extend(
                                    ["--fastq-paired", str(r1_path), str(r1_read_len), str(r2_path), str(r2_read_len)]
                                )
                            checked_files.add(r1_path)
                            checked_files.add(r2_path)
                else:
                    for f_meta in fastq_files:
                        f_path = self.files_dir / f_meta.file_path
                        if f_path in checked_files:
                            continue
                        if should_check_file(f_path, f_meta):
                            read_len = -1  # disable read length check for now; f_meta.read_length
                            grz_check_args.extend(["--fastq-single", str(f_path), str(read_len)])
                        checked_files.add(f_path)

                for bam_meta in bam_files:
                    bam_path = self.files_dir / bam_meta.file_path
                    if bam_path in checked_files:
                        continue
                    if should_check_file(bam_path, bam_meta):
                        grz_check_args.extend(["--bam", str(bam_path)])
                    checked_files.add(bam_path)

        # Handle any other files with --raw for calculating checksums
        for file_path, file_metadata in self.files.items():
            if file_path not in checked_files and should_check_file(file_path, file_metadata):
                grz_check_args.extend(["--raw", str(file_path)])

        if not grz_check_args:
            self.__log.info("All files are already validated. Skipping `grz-check`.")
        else:
            log_dir = Path(checksum_progress_file).parent
            temp_report_path = log_dir / "grz-check.report.jsonl"
            temp_report_path.unlink(missing_ok=True)

            command_args = ["--output", str(temp_report_path), *grz_check_args]
            if threads:
                command_args.extend(["--threads", str(threads)])
            try:
                run_grz_check(command_args)
            except UserInterruptException:
                self.__log.warning("Validation cancelled by user. Processing partial results...")
                raise
            except subprocess.CalledProcessError as e:
                self.__log.error(f"`grz-check` failed with exit code {e.returncode}")
                yield "`grz-check` execution failed. See logs for details."
            finally:
                if temp_report_path.is_file():
                    self.__log.info(f"Processing report file: {temp_report_path}")
                    with temp_report_path.open("r") as f:
                        self._process_grz_check_report(f, checksum_progress_logger, seq_data_progress_logger)  # type: ignore
                    temp_report_path.unlink()
                else:
                    self.__log.warning("`grz-check` did not produce a report file.")

        # Aggregate errors from both logs
        all_errors = set()
        for local_file_path, file_metadata in self.files.items():
            checksum_state = checksum_progress_logger.get_state(local_file_path, file_metadata)
            if checksum_state and not checksum_state.get("validation_passed"):
                for error in checksum_state.get("errors", []):
                    all_errors.add(f"{local_file_path.relative_to(self.files_dir)}: {error}")

            if file_metadata.file_type in ("fastq", "bam"):
                seq_data_state = seq_data_progress_logger.get_state(local_file_path, file_metadata)
                if seq_data_state and not seq_data_state.get("validation_passed"):
                    for error in seq_data_state.get("errors", []):
                        all_errors.add(f"{local_file_path.relative_to(self.files_dir)}: {error}")
        yield from all_errors

    def _process_grz_check_report(  # noqa: C901
        self,
        report_file: typing.TextIO,
        checksum_progress_logger: FileProgressLogger[ValidationState],
        seq_data_progress_logger: FileProgressLogger[ValidationState],
    ):
        """
        Parses the JSONL report from `grz-check` and updates both progress loggers.
        """
        for line in report_file:
            try:
                report_entry = json.loads(line)
                data = report_entry.get("data", {})
                file_path_str = data.get("path")
                if not file_path_str:
                    continue

                file_path = Path(file_path_str).resolve()
                file_metadata = self.files.get(file_path)

                if not file_metadata:
                    self.__log.warning(f"Could not find metadata for file in grz-check report: {file_path_str}")
                    continue

                status = data.get("status")
                errors = data.get("errors", [])
                warnings = data.get("warnings", [])
                checksum = data.get("checksum")

                if warnings:
                    for w in warnings:
                        self.__log.warning(w)

                checksum_issues = []
                if (
                    checksum
                    and (file_metadata.checksum_type or "").lower() == "sha256"
                    and file_metadata.file_checksum != checksum
                ):
                    checksum_issues.append(
                        f"Checksum mismatch! Expected: '{file_metadata.file_checksum}', calculated: '{checksum}'"
                    )

                if file_path.exists() and file_path.is_file():
                    if file_metadata.file_size_in_bytes != file_path.stat().st_size:
                        checksum_issues.append(
                            f"File size mismatch! Expected: '{file_metadata.file_size_in_bytes}', observed: '{file_path.stat().st_size}'."
                        )
                else:
                    checksum_issues.append("File not found for size check.")

                checksum_passed = not checksum_issues
                checksum_state = ValidationState(errors=checksum_issues, validation_passed=checksum_passed)
                checksum_progress_logger.set_state(file_path, file_metadata, checksum_state)

                if file_metadata.file_type in ("fastq", "bam"):
                    integrity_passed = status == "OK"
                    integrity_errors = errors if not integrity_passed else []

                    seq_data_state = ValidationState(errors=integrity_errors, validation_passed=integrity_passed)
                    seq_data_progress_logger.set_state(file_path, file_metadata, seq_data_state)

            except json.JSONDecodeError:
                self.__log.warning(f"Could not parse line in grz-check report: {line.strip()}")
            except Exception as e:
                self.__log.error(f"Error processing grz-check report entry: {line.strip()}. Error: {e}")

    @staticmethod
    def _validate_file_data_fallback(metadata: File, local_file_path: Path) -> Generator[str]:
        """
        Validates whether the provided file matches this metadata.
        (Fallback method)

        :param metadata: Metadata model object
        :param local_file_path: Path to the actual file (resolved if symlinked)
        :return: Generator of errors
        """
        # Resolve file path
        local_file_path = local_file_path.resolve()

        # Check if path exists
        if not local_file_path.exists():
            yield f"{str(Path('files') / metadata.file_path)} does not exist! Ensure filePath is relative to the files/ directory under the submission root."
            # Return here as following tests cannot work
            return

        # Check if path is a file
        if not local_file_path.is_file():
            yield f"{str(metadata.file_path)} is not a file!"
            # Return here as following tests cannot work
            return

        # Check if the checksum is correct
        if metadata.checksum_type == "sha256":
            calculated_checksum = calculate_sha256(local_file_path)
            if metadata.file_checksum != calculated_checksum:
                yield (
                    f"{str(metadata.file_path)}: Checksum mismatch! "
                    f"Expected: '{metadata.file_checksum}', calculated: '{calculated_checksum}'."
                )
        else:
            yield (
                f"{str(metadata.file_path)}: Unsupported checksum type: {metadata.checksum_type}. "
                f"Supported types: {[e.value for e in ChecksumType]}"
            )

        # Check file size
        if metadata.file_size_in_bytes != local_file_path.stat().st_size:
            yield (
                f"{str(metadata.file_path)}: File size mismatch! "
                f"Expected: '{metadata.file_size_in_bytes}', observed: '{local_file_path.stat().st_size}'."
            )

    def _validate_checksums_fallback(self, progress_log_file: str | PathLike) -> Generator[str]:
        """
        Validates the checksum of the files against the metadata.
        (Fallback method)

        :return: Generator of errors
        """
        progress_logger = FileProgressLogger[ValidationState](log_file_path=progress_log_file)
        # cleanup log file and keep only files listed here
        progress_logger.cleanup(keep=[(file_path, file_metadata) for file_path, file_metadata in self.files.items()])
        # fields:
        # - "errors": List[str]
        # - "validation_passed": bool

        def validate_file(local_file_path, file_metadata):
            self.__log.debug("Validating '%s'...", str(local_file_path))

            # validate the file
            errors = list(self._validate_file_data_fallback(file_metadata, local_file_path))
            validation_passed = len(errors) == 0

            # return log state
            return ValidationState(errors=errors, validation_passed=validation_passed)

        for local_file_path, file_metadata in self.files.items():
            logged_state = progress_logger.get_state(
                local_file_path,
                file_metadata,
                default=validate_file,  # validate the file if the state was not calculated yet
            )

            if logged_state:
                yield from logged_state["errors"]

    def _validate_sequencing_data_fallback(self, progress_log_file: str | PathLike) -> Generator[str]:
        """
        Quick-validates sequencing data linked in this submission.
        (Fallback method)

        :return: Generator of errors
        """
        from ..progress import FileProgressLogger

        progress_logger = FileProgressLogger[ValidationState](log_file_path=progress_log_file)
        # cleanup log file and keep only files listed here
        progress_logger.cleanup(keep=[(file_path, file_metadata) for file_path, file_metadata in self.files.items()])
        # fields:
        # - "errors": List[str]
        # - "validation_passed": bool

        def find_fastq_files(sequence_data: SequenceData) -> list[File]:
            return [f for f in sequence_data.files if f.file_type == FileType.fastq]

        def find_bam_files(sequence_data: SequenceData) -> list[File]:
            return [f for f in sequence_data.files if f.file_type == FileType.bam]

        for donor in self.metadata.content.donors:
            for lab_data in donor.lab_data:
                sequencing_layout = lab_data.sequencing_layout
                sequence_data = lab_data.sequence_data
                # find all FASTQ files
                fastq_files = find_fastq_files(sequence_data) if sequence_data else []
                bam_files = find_bam_files(sequence_data) if sequence_data else []

                if not lab_data.library_type.endswith("_lr"):
                    match sequencing_layout:
                        case SequencingLayout.single_end | SequencingLayout.reverse | SequencingLayout.other:
                            yield from self._validate_single_end_fallback(fastq_files, progress_logger)
                        case SequencingLayout.paired_end:
                            yield from self._validate_paired_end_fallback(fastq_files, progress_logger)
                yield from self._validate_bams_fallback(bam_files, progress_logger)

    def _validate_bams_fallback(
        self,
        bam_files: list[File],
        progress_logger: FileProgressLogger[ValidationState],
    ) -> Generator[str, None, None]:
        """
        Basic BAM sanity checks.
        (Fallback method)

        :param bam_files: List of BAM files
        :param progress_logger: Progress logger
        """

        def validate_file(local_file_path, _file_metadata) -> ValidationState:
            self.__log.debug("Validating '%s'...", str(local_file_path))

            # validate the file
            errors = list(validate_bam(local_file_path))
            validation_passed = len(errors) == 0

            # return log state
            return ValidationState(
                errors=errors,
                validation_passed=validation_passed,
            )

        for bam_file in bam_files:
            logged_state = progress_logger.get_state(
                self.files_dir / bam_file.file_path,
                bam_file,
                default=validate_file,  # validate the file if the state was not calculated yet
            )
            if logged_state:
                yield from logged_state["errors"]

    def _validate_single_end_fallback(
        self,
        fastq_files: list[File],
        progress_logger: FileProgressLogger[ValidationState],
    ) -> Generator[str, None, None]:
        def validate_file(local_file_path, file_metadata: SubmissionFileMetadata) -> ValidationState:
            self.__log.debug("Validating '%s'...", str(local_file_path))

            # validate the file
            errors = list(validate_single_end_reads(local_file_path, expected_read_length=file_metadata.read_length))
            validation_passed = len(errors) == 0

            # return log state
            return ValidationState(errors=errors, validation_passed=validation_passed)

        for fastq_file in fastq_files:
            logged_state = progress_logger.get_state(
                self.files_dir / fastq_file.file_path,
                fastq_file,
                default=validate_file,  # validate the file if the state was not calculated yet
            )
            if logged_state:
                yield from logged_state["errors"]

    def _validate_paired_end_fallback(
        self,
        fastq_files: list[File],
        progress_logger: FileProgressLogger[ValidationState],
    ) -> Generator[str, None, None]:
        key = lambda f: (f.flowcell_id, f.lane_id)
        fastq_files.sort(key=key)
        for _key, group in groupby(fastq_files, key):
            files = list(group)

            # separate R1 and R2 files
            fastq_r1_files = [f for f in files if f.read_order == ReadOrder.r1]
            fastq_r2_files = [f for f in files if f.read_order == ReadOrder.r2]

            for fastq_r1, fastq_r2 in zip(fastq_r1_files, fastq_r2_files, strict=True):
                local_fastq_r1_path = self.files_dir / fastq_r1.file_path
                local_fastq_r2_path = self.files_dir / fastq_r2.file_path

                # get saved state
                logged_state_r1 = progress_logger.get_state(local_fastq_r1_path, fastq_r1)
                logged_state_r2 = progress_logger.get_state(local_fastq_r2_path, fastq_r2)

                if logged_state_r1 is None or logged_state_r2 is None or logged_state_r1 != logged_state_r2:
                    # calculate state
                    errors = list(
                        validate_paired_end_reads(
                            local_fastq_r1_path,  # fastq R1
                            local_fastq_r2_path,  # fastq R2
                            expected_read_length=fastq_r1.read_length,  # fastq R1 and R2 should have the same read length
                        )
                    )
                    validation_passed = len(errors) == 0

                    state = ValidationState(errors=errors, validation_passed=validation_passed)
                    # update state for both files
                    progress_logger.set_state(  # fastq R1
                        local_fastq_r1_path, fastq_r1, state
                    )
                    progress_logger.set_state(  # fastq R2
                        local_fastq_r2_path, fastq_r2, state
                    )
                    yield from state["errors"]
                else:
                    # both fastq states are equal, so simply yield one of them
                    yield from logged_state_r1["errors"]

    def encrypt(
        self,
        encrypted_files_dir: str | PathLike,
        progress_log_file: str | PathLike,
        recipient_public_key_path: str | PathLike,
        submitter_private_key_path: str | PathLike | None = None,
        force: bool = False,
    ) -> EncryptedSubmission:
        """
        Encrypt this submission with a public key using Crypt4Gh

        :param encrypted_files_dir: Output directory of the encrypted files
        :param progress_log_file: Path to a log file to store the progress of the encryption process
        :param recipient_public_key_path: Path to the public key file which will be used for encryption
        :param submitter_private_key_path: Path to the private key file which will be used to sign the encryption
        :param force: Force encryption even if target files already exist
        :return: EncryptedSubmission instance
        """
        encrypted_files_dir = Path(encrypted_files_dir)

        if not Path(recipient_public_key_path).expanduser().is_file():
            msg = f"Public key file does not exist: {recipient_public_key_path}"
            self.__log.error(msg)
            raise FileNotFoundError(msg)
        if not submitter_private_key_path:
            self.__log.warning("No submitter private key provided, skipping signing.")
        elif not Path(submitter_private_key_path).expanduser().is_file():
            msg = f"Private key file does not exist: {submitter_private_key_path}"
            self.__log.error(msg)
            raise FileNotFoundError(msg)

        if not encrypted_files_dir.is_dir():
            self.__log.debug(
                "Creating encrypted submission files directory: %s...",
                encrypted_files_dir,
            )
            encrypted_files_dir.mkdir(mode=0o770, parents=False, exist_ok=False)

        from ..progress import FileProgressLogger

        progress_logger = FileProgressLogger[EncryptionState](log_file_path=progress_log_file)

        try:
            public_keys = Crypt4GH.prepare_c4gh_keys(recipient_public_key_path)
        except Exception as e:
            self.__log.error(f"Error preparing public keys: {e}")
            raise e

        for file_path, file_metadata in self.files.items():
            # encryption_successful = True
            logged_state = progress_logger.get_state(file_path, file_metadata)
            self.__log.debug("state for %s: %s", file_path, logged_state)

            encrypted_file_path = encrypted_files_dir / EncryptedSubmission.get_encrypted_file_path(
                file_metadata.file_path
            )
            encrypted_file_path.parent.mkdir(mode=0o770, parents=True, exist_ok=True)

            if (
                (logged_state is None)
                or not logged_state.get("encryption_successful", False)
                or not encrypted_file_path.is_file()
            ):
                self.__log.info(
                    "Encrypting file: '%s' -> '%s'",
                    str(file_path),
                    str(encrypted_file_path),
                )

                if encrypted_file_path.exists() and not force:
                    raise RuntimeError(
                        f"'{encrypted_file_path}' already exists. Delete it or use --force to overwrite it."
                    )

                try:
                    Crypt4GH.encrypt_file(file_path, encrypted_file_path, public_keys)

                    self.__log.info(f"Encryption complete for {str(file_path)}. ")
                    progress_logger.set_state(
                        file_path,
                        file_metadata,
                        state=EncryptionState(encryption_successful=True),
                    )
                except Exception as e:
                    self.__log.error("Encryption failed for '%s'", str(file_path))

                    progress_logger.set_state(
                        file_path,
                        file_metadata,
                        state=EncryptionState(encryption_successful=False, errors=[str(e)]),
                    )

                    raise e
            else:
                self.__log.info(
                    "File '%s' already encrypted in '%s'",
                    str(file_path),
                    str(encrypted_file_path),
                )

        self.__log.info("File encryption completed.")

        return EncryptedSubmission(
            metadata_dir=self.metadata_dir,
            encrypted_files_dir=encrypted_files_dir,
        )


class EncryptedSubmission:
    """The encrypted counterpart to `Submission`. Handles encrypted submission data."""

    __log = log.getChild("EncryptedSubmission")

    def __init__(
        self, metadata_dir: str | PathLike, encrypted_files_dir: str | PathLike, log_dir: str | PathLike | None = None
    ):
        """
        Initialize the encrypted submission object.

        :param metadata_dir: Path to the metadata directory
        :param encrypted_files_dir: Path to the encrypted files directory
        """
        self.metadata_dir = Path(metadata_dir)
        self.encrypted_files_dir = Path(encrypted_files_dir)
        self.log_dir = Path(log_dir) if log_dir is not None else None

        self.metadata = SubmissionMetadata(self.metadata_dir / "metadata.json")

    @property
    def encrypted_files(self) -> dict[Path, SubmissionFileMetadata]:
        """
        The encrypted files linked in the metadata.

        :return: Dictionary of `local_file_path` -> `SubmissionFileMetadata` pairs.
        """
        retval = {}
        for file_path, file_metadata in self.metadata.files.items():
            encrypted_file_path = self.get_encrypted_file_path(self.encrypted_files_dir / file_path)

            retval[encrypted_file_path] = file_metadata

        return retval

    @property
    def submission_id(self) -> str:
        return self.metadata.content.submission_id

    def get_metadata_file_path_and_object_id(self) -> tuple[Path, str]:
        """
        :return: tuple with the `local_file_path` and s3_object_id of the metadata file
        """
        return Path(self.metadata.file_path), str(Path(self.submission_id) / "metadata" / self.metadata.file_path.name)

    def get_encrypted_files_and_object_id(self) -> dict[Path, str]:
        """
        :return Dictionary of `local_file_path` -> s3_object_id
        """
        retval = {}
        for local_file_path, file_metadata in self.encrypted_files.items():
            retval[local_file_path] = str(
                Path(self.submission_id) / "files" / self.get_encrypted_file_path(file_metadata.file_path)
            )
        return retval

    def get_log_files_and_object_id(self) -> dict[Path, str]:
        """
        :return Dictionary of `local_file_path` -> s3_object_id
        """
        retval = {}
        if self.log_dir is not None:
            log_dir = self.log_dir
            for dirpath, _dirnames, filenames in log_dir.walk():
                for filename in filenames:
                    local_file_path = dirpath / filename
                    retval[local_file_path] = str(
                        Path(self.submission_id) / "logs" / local_file_path.relative_to(log_dir)
                    )
        return retval

    @staticmethod
    def get_encrypted_file_path(file_path: str | PathLike) -> Path:
        """
        Return the path to the encrypted file based on the original file path,
        with additional extension'.c4gh'.
        """
        p = Path(file_path)
        return p.with_suffix(p.suffix + ".c4gh")

    @staticmethod
    def get_encryption_header_path(file_path: str | PathLike) -> Path:
        """
        Return the path to the encryption header file based on the original file path,
        with additional extension'.c4gh_header'.
        """
        p = Path(file_path)
        return p.with_suffix(p.suffix + ".c4gh_header")

    def decrypt(
        self,
        files_dir: str | PathLike,
        progress_log_file: str | PathLike,
        recipient_private_key_path: str | PathLike,
    ) -> Submission:
        """
        Decrypt this encrypted submission with a private key using Crypt4Gh

        :param files_dir: Output directory of the decrypted files
        :param progress_log_file: Path to a log file to store the progress of the decryption process
        :param recipient_private_key_path: Path to the private key file which will be used for decryption
        :return: Submission instance
        """
        files_dir = Path(files_dir)

        if not files_dir.is_dir():
            self.__log.debug(
                "Creating decrypted submission files directory: %s...",
                files_dir,
            )
            files_dir.mkdir(mode=0o770, parents=False, exist_ok=False)

        from ..progress import FileProgressLogger

        progress_logger = FileProgressLogger[DecryptionState](log_file_path=progress_log_file)

        try:
            private_key = Crypt4GH.retrieve_private_key(recipient_private_key_path)
        except Exception as e:
            self.__log.error(f"Error preparing private key: {e}")
            raise e

        for encrypted_file_path, file_metadata in self.encrypted_files.items():
            logged_state = progress_logger.get_state(encrypted_file_path, file_metadata)
            self.__log.debug("state for %s: %s", encrypted_file_path, logged_state)

            decrypted_file_path = files_dir / file_metadata.file_path
            if not decrypted_file_path.parent.is_dir():
                decrypted_file_path.parent.mkdir(mode=0o770, parents=True, exist_ok=False)

            if (
                (logged_state is None)
                or not logged_state.get("decryption_successful", False)
                or not decrypted_file_path.is_file()
            ):
                self.__log.info(
                    "Decrypting file: '%s' -> '%s'",
                    str(encrypted_file_path),
                    str(decrypted_file_path),
                )

                try:
                    Crypt4GH.decrypt_file(encrypted_file_path, decrypted_file_path, private_key)

                    self.__log.info(f"Decryption complete for {str(encrypted_file_path)}. ")
                    progress_logger.set_state(
                        encrypted_file_path,
                        file_metadata,
                        state=DecryptionState(decryption_successful=True),
                    )
                except Exception as e:
                    self.__log.error("Decryption failed for '%s'", str(encrypted_file_path))

                    progress_logger.set_state(
                        encrypted_file_path,
                        file_metadata,
                        state=DecryptionState(decryption_successful=False, errors=[str(e)]),
                    )

                    raise e
            else:
                self.__log.info(
                    "File '%s' already decrypted in '%s'",
                    str(encrypted_file_path),
                    str(decrypted_file_path),
                )

        self.__log.info("File decryption completed.")

        return Submission(
            metadata_dir=self.metadata_dir,
            files_dir=files_dir,
        )


class SubmissionValidationError(Exception):
    """Exception raised when validation of a submission fails"""

    pass
